"""
.. codeauthor:: Nathan Hertz <nhertz@nrao.edu>

Workspaces testing utilities

Put things here if you'd like them to be available to any given test in the entire system

HOW TO USE:
1. In any given pytest-powered test, add this line below the imports:
pytest_plugins = ["testing.utils.conftest"]
2. All fixtures within this file will now be available for use
3. If you want a non-fixture item from here, you'll have to do a normal Python import
"""
from typing import List
from unittest import mock
from unittest.mock import patch

import pytest
from mock_alchemy.mocking import UnifiedAlchemyMagicMock

from workspaces.capability.schema import (
    Capability,
    CapabilityRequest,
    CapabilityVersion,
)
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityRequestIF,
)
from workspaces.capability.services.capability_engine import CapabilityEngine
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.capability_service import CapabilityService
from workspaces.capability.services.interfaces import CapabilityEngineIF
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import WorkflowService


@pytest.fixture(scope="session")
def mock_capability_info() -> CapabilityInfo:
    """
    Pytest fixture providing a mock CapabilityInfo object with a test capability; mocked object can be used
    like a normal CapabilityInfo instance
    TODO: Use UnifiedAlchemyMagicMock's data parameter to add test capabilities

    :return: Mocked CapabilityInfo
    """
    null_capability = Capability(
        name="null",
        steps="prepare-and-run-workflow null\nawait-workflow",
        max_jobs=2,
    )
    test_capability = Capability(name="test", steps="await-qa", max_jobs=777)
    null_capability_query = (
        [mock.call.query(Capability), mock.call.filter_by(name="null")],
        [null_capability],
    )
    test_capability_query = (
        [mock.call.query(Capability), mock.call.filter_by(name="test")],
        [test_capability],
    )
    null_capability_request = CapabilityRequest(
        id=1,
        state="Ready",
        capability=null_capability,
        parameters="-g",
        # a trick here is to ensure that we always have a first version, with the original parameters
        versions=[CapabilityVersion(version_number=1, parameters="-g")],
    )
    null_capability_request_query = (
        [mock.call.query(CapabilityRequest), mock.call.filter_by(id=1)],
        [null_capability_request],
    )
    mocked_capability_info = CapabilityInfo(
        UnifiedAlchemyMagicMock(
            data=[null_capability_query, test_capability_query, null_capability_request_query],
        )
    )

    return mocked_capability_info


@pytest.fixture(scope="session")
def mock_capability_service(
    mock_capability_info: CapabilityInfo,
    mock_workflow_service: WorkflowService,
) -> CapabilityService:
    """
    Pytest fixture providing a mock CapabilityService object with no threaded event listener creation

    :return: Mocked CapabilityService
    """
    with patch("threading.Thread"):
        with patch("threading.Thread.start"):
            yield CapabilityService(mock_capability_info, mock_workflow_service)


@pytest.fixture(scope="session")
def mock_workflow_info() -> WorkflowInfo:
    """
    Pytest fixture providing a mock WorkflowInfo object; mocked object can be used
    like a normal WorkflowInfo instance

    :return: Mocked WorkflowInfo
    """
    return WorkflowInfo(UnifiedAlchemyMagicMock())


@pytest.fixture(scope="session")
def mock_workflow_service(mock_workflow_info: WorkflowInfo) -> WorkflowService:
    """
    Pytest fixture providing a mock WorkflowService object

    :return: Mocked WorkflowService
    """
    with patch("threading.Thread"):
        with patch("threading.Thread.start"):
            yield WorkflowService(mock_workflow_info)


@pytest.fixture(scope="session")
def mock_capability_request(
    mock_capability_service: CapabilityService,
) -> List[CapabilityRequestIF]:
    """
    Pytest fixture adding two mock CapabilityRequest objects that can be accessed from the mock_capability_info

    :param mock_capability_service: Mocked CapabilityService
    :return: Mocked CapabilityRequest
    """
    requests = []
    r1 = mock_capability_service.create_request("null", ["-g"])
    r1.id = 0
    requests.append(r1)
    r2 = mock_capability_service.create_request("test")
    r2.id = 1
    requests.append(r2)
    return requests


@pytest.fixture(scope="session")
def mock_capability_execution(
    mock_capability_info: CapabilityInfo, mock_capability_request: CapabilityRequest
) -> CapabilityExecutionIF:
    """
    Pytest fixture adding a mock CapabilityExecution object that can be
    accessed from mock_capability_info

    :param mock_capability_request: Mocked CapabilityRequest
    :param mock_capability_info: Mocked CapabilityInfo
    :return: Mocked CapabilityExecution
    """
    request = mock_capability_request[0]

    execution = mock_capability_info.create_execution(
        mock_capability_info.lookup_capability_request(request)
    )
    return execution


@pytest.fixture(scope="session")
def mock_capability_engine() -> CapabilityEngineIF:
    """
    Pytest fixture adding a mock CapabilityEngine object
    :return: mocked engine
    """

    return CapabilityEngine(mock_capability_info, mock_workflow_service)
